/*
 * Copyright (c) 2003-2011 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <netdb.h>
#include <poll.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>

#include "network.h"
#include "serial.h"

extern int optind;
static int verbose = 0;

/* Timeout value before re-contacting proxy if not data arrive */
#define DGPS_TIMEOUT 60000 /* 60s */

static void
usage(const char *progname)
{
	fprintf(stderr, "writes data received on UDP port %d to serial port\n", GPS_PORT);
	fprintf(stderr, "the GPS device must be configured to receive data\n");
	fprintf(stderr, "at 57600 bauds, 8bits, no parity, one bit stop\n");
	fprintf(stderr, "usage: %s [-v][-f][-n][-p proxy] serial_port\n", 
	    progname);
	fprintf(stderr, "  -v (verbose) print to the screen the received data\n");
	fprintf(stderr, "  -f don't configure the serial port before opening it\n");
	fprintf(stderr, "  -n  don't write on the serial port\n");
	fprintf(stderr, "  -p proxy use proxy server to get corrections\n");
	exit(1);
}

/*
 * Connect to the proxy server
 */
static int
register_proxy(char *proxy)
{
	GPS_SEND_STR *proxyConn;
	char proxyHello[] = DGPS_PROXY_HELLO;
	size_t nSent;
	int fd;
	
	proxyConn = setupToSocket(proxy, GPS_PROXY_PORT);
	if (proxyConn == NULL) 
		return -1;
	nSent = sendto(proxyConn->fd, proxyHello, sizeof(proxyHello),
	    0, (struct sockaddr *)&proxyConn->dst, proxyConn->len);
	if (nSent != sizeof(proxyHello))
		return -1;

	fd = proxyConn->fd;
	free(proxyConn);
	return fd;
}

/*
 * Try to contact the proxy when needed:
 *
 * - if there is no connection yet
 * - if no packet has been received on the existing connection for GPS_TIMEOUT
 *
 * When this returns, if inFd >= 0, then there are data available
 */
static void
check_proxy(char *proxy, int *inFd)
{
	struct pollfd fds;
	struct timespec ts;
			
	if (*inFd == -1) {
		/* Need to connect to the proxy */
		if (verbose) 
			printf("(Re)contacting proxy %s\n", 
			    proxy);
		*inFd = register_proxy(proxy);
	}
	if (*inFd == -1) {
		/* Connection failed, sleep for a while before retrying */
		ts.tv_sec = 10;
		ts.tv_nsec = 0;
		nanosleep(&ts, NULL);
		return;
	}
	/* a proxy connection exists, check that data is received 
	 * before timeout */
	fds.fd = *inFd;
	fds.events = POLLIN;
	if (poll(&fds, 1, DGPS_TIMEOUT) == 0) {
		/* timeout - re-register */
		if (verbose)
			printf("Timeout on receive.\n");
		close(*inFd);
		*inFd = -1;
	}
}

int
main(int argc, char *argv[])
{
	int ch, inFd = -1, outFd, i, n, nWritten;
	socklen_t len = sizeof(struct sockaddr_in);
	struct sockaddr_in from;
	char buf[1500];
	char *prog = argv[0];
	char *proxy = NULL;
	int useSerial = 1;
	int nowrite = 0;

	while ((ch = getopt(argc, argv, "fnp:v")) != -1) {
		switch (ch) {
		case 'v':
			verbose = 1;
			break;
		case 'n':
			nowrite = 1;
			break;
		case 'f':
			useSerial = 0;
			break;
		case 'p':
			proxy = optarg;
			break;
		case '?':
		default:
			usage(prog);
		} /* switch */
	}
	argc -= optind;
	argv += optind;

	if ((!nowrite && argc != 1) || (nowrite && argc > 1)) {
		usage(prog);
	}

	if (!nowrite)
	{
		if (useSerial)
				outFd = openSerial(argv[0]);
		else
				outFd = open(argv[0], O_CREAT | O_WRONLY, 0644);
		
		if (outFd < 0) {
			exit(1);
		}
	}

	if (proxy == NULL) {
		inFd = setupRcvSocket(GPS_PORT);
		if (inFd < 0)
			exit(1);
	}
	while (1) {
		/* Handle proxy connection if needed */
		if (proxy != NULL) {
			check_proxy(proxy, &inFd);
			if (inFd == -1)
				continue;
		}
		n = recvfrom(inFd, buf, sizeof(buf), 0, 
		    (struct sockaddr *)&from, &len);
		if (n < 0) {
			fprintf(stderr, "%s:%d: recvfrom: %s\n", __FILE__,
			    __LINE__, strerror(errno));
			break;
		}
		if (verbose) {
			printf("received %d bytes from %s\n", n, 
			    inet_ntoa(from.sin_addr));
			for (i = 0; i < n; i++) {
				printf("%#02x ", buf[i]);
			}
			printf("\n");
		}
		if (nowrite) 
			continue;
		
		if (useSerial)
			nWritten= writeSerial(outFd, buf, n);
		else
			nWritten = write(outFd, buf, n);
		
		if (nWritten != n) {
			fprintf(stderr, "%s:%d: writeSerial %d != %d: %s\n", 
			    __FILE__, __LINE__, nWritten, n, strerror(errno));
		}
	}

	close(inFd);
	if (!nowrite) 
		close(outFd);
	exit(0);
}
		
